@@ -1,14 +1,14 @@
 #  You may distribute under the terms of either the GNU General Public License
 #  or the Artistic License (the same terms as Perl itself)
 #
-#  (C) Paul Evans, 2014-2017 -- leonerd@leonerd.org.uk
+#  (C) Paul Evans, 2014-2019 -- leonerd@leonerd.org.uk
 
 package Data::Bitfield;
 
 use strict;
 use warnings;
 
-our $VERSION = '0.02';
+our $VERSION = '0.03';
 
 use Exporter 'import';
 our @EXPORT_OK = qw( bitfield boolfield intfield enumfield constfield );
@@ -17,28 +17,63 @@ use Carp;
 
 =head1 NAME
 
-C<Data::Bitfield> - manage integers containing multiple bit fields
+C<Data::Bitfield> - manage data packed into bytes of multiple bit fields
+
+=head1 SYNOPSIS
+
+   use Data::Bitfield qw( bitfield boolfield enumfield );
+
+   # The stat(2) st_mode field on Linux
+   bitfield MODE =>
+      format      => enumfield(12,
+         undef,     "fifo", "char", undef, "dir",    undef, "block", undef,
+         "regular", undef,  "link", undef, "socket", undef, undef,   undef ),
+      set_uid     => boolfield(11),
+      set_gid     => boolfield(10),
+      sticky      => boolfield(9),
+      user_read   => boolfield(8),
+      user_write  => boolfield(7),
+      user_exec   => boolfield(6),
+      group_read  => boolfield(5),
+      group_write => boolfield(4),
+      group_exec  => boolfield(3),
+      other_read  => boolfield(2),
+      other_write => boolfield(1),
+      other_exec  => boolfield(0);
+
+   my %modebits = unpack_MODE( stat($path)->mode );
+
+Z<>
+
+   # The flag register of a Z80
+   bitfield FLAGS =>
+      sign      => boolfield(7),
+      zero      => boolfield(6),
+      halfcarry => boolfield(4),
+      parity    => boolfield(2),
+      subtract  => boolfield(1),
+      carry     => boolfield(0);
 
 =head1 DESCRIPTION
 
 This module provides a single primary function, C<bitfield>, which creates
-helper functions in the package that calls it, to assist in managing integers
-that encode sets of bits, called bitfields. This may be useful when
+helper functions in the package that calls it, to assist in managing data that
+is encoded in sets of bits, called bitfields. This may be useful when
 interacting with a low-level networking protocol, binary file format, hardware
 devices, or similar purposes.
 
 =head2 bitfield
 
- bitfield $name, %fields
+   bitfield $name, %fields
 
 Creates two new functions in the calling package whose names are derived from
 the string C<$name> passed here. These functions will be symmetric opposites,
 which convert between a key/value list of field values, and their packed
-integer or binary byte-string representation.
+binary byte-string or integer representation.
 
- $packed_value = pack_$name( %field_values )
+   $packed_value = pack_$name( %field_values )
 
- %field_values = unpack_$name( $packed_value )
+   %field_values = unpack_$name( $packed_value )
 
 These two functions will work to a set of field names that match those field
 definitions given to the C<bitfield> function that declared them.
@@ -49,7 +84,7 @@ following field-declaration functions.
 Additional options may be passed by giving a C<HASH> reference as the first
 argument, before the structure name.
 
- bitfield { %options }, $name, %fields
+   bitfield { %options }, $name, %fields
 
 Recognised options are:
 
@@ -62,8 +97,8 @@ C<unpack_NAME> function will expect to receive as input. The two C<bytes-*>
 formats describe a packed binary string in either little- or big-endian
 direction, and C<integer> describes an integer numerical value.
 
-Note that currently the C<bytes-*> formats are limited to values 32bits wide
-or smaller.
+Note that currently the C<integer> format is limited to values 32bits wide or
+smaller.
 
 Optional; will default to C<integer> if not supplied. This default may change
 in a later version - make sure to always specify it for now.
@@ -96,128 +131,127 @@ sub bitfield_into_caller
    $VALID_FORMATS{$format} or
       croak "Invalid 'format' value $format";
 
-   my $used_bits = 0;
+   my $used_bits = '';
 
-   my $constmask = 0;
-   my $constval = 0;
+   my $constmask = '';
+   my $constval = '';
+
+   my %fields;
 
-   my %fieldmask;
-   my %fieldshift;
-   my %fieldencoder;
-   my %fielddecoder;
    while( @args ) {
       my $name = shift @args;
       if( !defined $name ) {
-         my ( $mask, $value ) = @{ shift @args };
+         my ( $shift, $width, $value ) = @{ shift @args };
+         my $mask = pack "L<", ( ( 1 << $width ) - 1 ) << $shift;
 
          croak "Constfield collides with other defined bits"
-            if $used_bits & $mask;
+            if ( $used_bits & $mask ) !~ m/^\0*$/;
 
-         $constmask |= $mask;
-         $constval |= $value;
+         $constval |= pack( "L<", $value << $shift );
          $used_bits |= $mask;
 
          next;
       }
 
-      ( my $mask, $fieldshift{$name}, $fieldencoder{$name}, $fielddecoder{$name} ) =
-         @{ shift @args };
+      my ( $shift, $width, $encoder, $decoder ) = @{ shift @args };
+      my $offs = int( $shift / 8 ); $shift %= 8;
+
+      my $mask = ( "\x00" x $offs ) . pack "L<", ( ( 1 << $width ) - 1 ) << $shift;
 
       croak "Field $name is defined twice"
-         if $fieldmask{$name};
+         if $fields{$name};
       croak "Field $name collides with other defined fields"
-         if $used_bits & $mask;
+         if ( $used_bits & $mask ) !~ m/^\0*$/;
 
-      $fieldmask{$name} = $mask;
+      $fields{$name} = [ $mask, $offs, $shift, $encoder, $decoder ];
       $used_bits |= $mask;
    }
 
-   my $nbits = 0;
-   $nbits += 8 while( 1 << $nbits < $used_bits );
+   $used_bits =~ s/\0+$//;
+   my $datalen = length $used_bits;
 
    my $packsub = sub {
       my %args = @_;
       my $ret = $constval;
       foreach ( keys %args ) {
-         my $mask = $fieldmask{$_};
-         next if !$mask and $unrecognised_ok;
-         croak "Unexpected field '$_'" unless $mask;
+         defined( my $f = $fields{$_} ) or
+            $unrecognised_ok and next or
+            croak "Unexpected field '$_'";
+
+         my ( $mask, $offs, $shift, $encoder ) = @$f;
 
          my $v = $args{$_};
-         $v = $fieldencoder{$_}($v) if $fieldencoder{$_};
+         $v = $encoder->($v) if $encoder;
          defined $v or croak "Unsupported value for '$_'";
 
-         if( defined( my $shift = $fieldshift{$_} ) ) {
+         {
             no warnings 'numeric';
-            int $v eq $v and $v >= 0 and $v <= ( $mask >> $shift ) or
+            int $v eq $v or
                croak "Expected an integer value for '$_'";
-            $ret |= ( $v << $shift ) & $mask;
-         }
-         else {
-            $ret |= $mask if $v;
          }
+
+         my $bits = ( "\x00" x $offs ) . ( pack "L<", $v << $shift );
+         $v >= 0 and ( $bits & ~$mask ) =~ m/^\0+$/ or
+            croak "Value out of range for '$_'";
+
+         $ret |= $mask & $bits;
       }
-      return $ret;
+      return substr( $ret, 0, $datalen );
    };
 
    my $unpacksub = sub {
       my ( $val ) = @_;
+      # Bitwise extend so there's always enough bits to unpack
+      $val .= "\0\0\0";
       # TODO: check constmask
       my @ret;
-      foreach ( keys %fieldmask ) {
-         my $v = $val & $fieldmask{$_};
-         if( defined( my $shift = $fieldshift{$_} ) ) {
-            $v >>= $shift;
-         }
-         else {
-            $v = !!$v;
-         }
+      foreach ( keys %fields ) {
+         my $f = $fields{$_};
+         my ( $mask, $offs, $shift, undef, $decoder ) = @$f;
 
-         $v = $fielddecoder{$_}($v) if $fielddecoder{$_};
+         my $v = unpack( "L<", substr( $val & $mask, $offs, 4 ) ) >> $shift;
+
+         $v = $decoder->($v) if $decoder;
          push @ret, $_ => $v;
       }
       return @ret;
    };
 
-   if( $format ne "integer" ) {
+   if( $format eq "bytes-BE" ) {
       my $orig_packsub   = $packsub;
       my $orig_unpacksub = $unpacksub;
 
-      my $big_endian = ( $format eq "bytes-BE" );
-      my $pointy = $big_endian ? ">" : "<";
+      $packsub = sub {
+         return scalar reverse $orig_packsub->(@_);
+      };
+      $unpacksub = sub {
+         return $orig_unpacksub->(scalar reverse $_[0]);
+      };
+   }
+   elsif( $format eq "integer" ) {
+      my $orig_packsub   = $packsub;
+      my $orig_unpacksub = $unpacksub;
+
+      my $nbits = $datalen * 8;
 
       if( $nbits <= 8 ) {
-         $packsub   = sub { pack "C", $orig_packsub->( @_ ) };
-         $unpacksub = sub { $orig_unpacksub->( unpack "C", $_[0] ) };
+         $packsub   = sub { unpack "C", $orig_packsub->( @_ ) };
+         $unpacksub = sub { $orig_unpacksub->( pack "C", $_[0] ) };
       }
       elsif( $nbits <= 16 ) {
-         $packsub   = sub { pack "S$pointy", $orig_packsub->( @_ ) };
-         $unpacksub = sub { $orig_unpacksub->( unpack "S$pointy", $_[0] ) };
+         $packsub   = sub { unpack "S<", $orig_packsub->( @_ ) };
+         $unpacksub = sub { $orig_unpacksub->( pack "S<", $_[0] ) };
       }
       elsif( $nbits <= 24 ) {
-         if( $big_endian ) {
-            $packsub = sub {
-               substr( pack( "L>", $orig_packsub->( @_ ) ), 1, 3 )
-            };
-            $unpacksub = sub {
-               $orig_unpacksub->( unpack "L>", "\0$_[0]" )
-            };
-         }
-         else {
-            $packsub = sub {
-               substr( pack( "L<", $orig_packsub->( @_ ) ), 0, 3 )
-            };
-            $unpacksub = sub {
-               $orig_unpacksub->( unpack "L<", "$_[0]\0" )
-            };
-         }
+         $packsub   = sub { unpack( "L<", $orig_packsub->( @_ ) . "\0" ) };
+         $unpacksub = sub { $orig_unpacksub->( pack "L<", $_[0] ) };
       }
       elsif( $nbits <= 32 ) {
-         $packsub   = sub { pack "L$pointy", $orig_packsub->( @_ ) };
-         $unpacksub = sub { $orig_unpacksub->( unpack "L$pointy", $_[0] ) };
+         $packsub   = sub { unpack "L<", $orig_packsub->( @_ ) };
+         $unpacksub = sub { $orig_unpacksub->( pack "L<", $_[0] ) };
       }
       else {
-         croak "Cannot currently handle bytewise packing of $nbits wide values";
+         croak "Cannot currently handle integer packing of $nbits wide values";
       }
    }
 
@@ -236,7 +270,7 @@ sub bitfield_into_caller
 
 =head2 boolfield
 
- boolfield( $bitnum )
+   boolfield( $bitnum )
 
 Declares a single bit-wide field at the given bit index, whose value is a
 simple boolean truth.
@@ -246,12 +280,12 @@ simple boolean truth.
 sub boolfield
 {
    my ( $bitnum ) = @_;
-   return [ 1 << $bitnum, undef ];
+   return [ $bitnum, 1, sub { 0 + !!shift }, sub { !!shift } ];
 }
 
 =head2 intfield
 
- intfield( $bitnum, $width )
+   intfield( $bitnum, $width )
 
 Declares a field of C<$width> bits wide, starting at the given bit index,
 whose value is an integer. It will be shifted appropriately.
@@ -261,13 +295,12 @@ whose value is an integer. It will be shifted appropriately.
 sub intfield
 {
    my ( $bitnum, $width ) = @_;
-   my $mask = ( 1 << $width ) - 1;
-   return [ $mask << $bitnum, $bitnum ];
+   return [ $bitnum, $width ];
 }
 
 =head2 enumfield
 
- enumfield( $bitnum, @values )
+   enumfield( $bitnum, @values )
 
 Declares a field some number of bits wide, sufficient to store an integer big
 enough to act as an index into the list of values, starting at the given bit
@@ -302,7 +335,7 @@ sub enumfield
 
 =head2 constfield
 
- constfield( $bitnum, $width, $value )
+   constfield( $bitnum, $width, $value )
 
 Declares a field some number of bits wide that stores a constant value. This
 value will be packed automatically.
@@ -319,8 +352,7 @@ sub constfield
    $value >= 0 and $value < ( 1 << $width ) or
       croak "Invalid value for constfield of width $width";
 
-   my $mask = ( 1 << $width ) - 1;
-   return undef, [ $mask << $bitnum, $value << $bitnum ];
+   return undef, [ $bitnum, $width, $value ];
 }
 
 =head1 TODO
